/*
 * $Id: lanwon.c,v 1.3 2003/06/21 21:30:48 felfert Exp $
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>
#include <getopt.h>
#include <syslog.h>
#include <signal.h>
#include <stdarg.h>

typedef struct query_s {
    char *name;
    unsigned char *query;
    unsigned char *reply;
    int query_len;
    int reply_check_len;
    int reply_check_offset;
    struct query_s *next;
} query_t;

typedef struct listener_s {
    int s;
    int port;
    struct listener_s *next;
    query_t *queries;
} listener_t;

static int verbose;
static int daemonized = 0;

static query_t bf1942_query = {
    "BF 1942 1.4",
    "\\status\\",
    "\\bfield1942\\",
    8,
    12,
    -1,
    NULL
};

static query_t bf1942_13_query = {
    "BF 1942 < 1.4",
    "\\status\\",
    "\\BF1942\\",
    8,
    8,
    -1,
    NULL
};

static query_t ut2003_query = {
    "UT 2003",
    "\170\000\000\000\000",
    "\170\000\000\000\000",
    5,
    5,
    0,
    NULL
};

static query_t ut_query = {
    "UT (old)",
    "REPORTQUERY",
    "7",
    11,
    1,
    4,
    NULL
};

static query_t hlds_details = {
    "Half-Life (Details)",
    "\377\377\377\377details\000",
    "m",
    12,
    1,
    4,
    NULL
};

static query_t hlds_infostring = {
    "Half-Life (Infostring)",
    "\377\377\377\377infostring\n\000",
    "i",
    16,
    1,
    4,
    NULL
};

static query_t q3_getinfo = {
    "Quake3",
    "\377\377\377\377getinfo xxx",
    "i",
    15,
    1,
    4,
    NULL
};

static void myperror(const char *prefix)
{
    if (daemonized)
	syslog(LOG_ERR, "%s: %m\n", prefix);
    else
	perror(prefix);
}

static int myprint(char *fmt, ...)
{
    va_list ap;
    int ret = 0;

    va_start(ap, fmt);
    if (!daemonized)
	ret = vfprintf(stdout, fmt, ap);
    else {
	char *buf;
	int len = vsnprintf(NULL, 0, fmt, ap) + 1;

	buf = (char *) malloc(len);
	if (buf) {
	    va_start(ap, fmt);
	    ret = vsnprintf(buf, len, fmt, ap);
	    syslog(LOG_INFO, "%s\n", buf);
	    free(buf);
	}
    }
    va_end(ap);
    return ret;
}

static int is_listening(struct in_addr *server, int sport)
{
    FILE *f;
    int ret = 0;
    char line[4096];

    f = fopen("/proc/net/udp", "r");
    if (!f) {
	myperror("/proc/net/udp");
	exit(-1);
    }
    // skip header
    fgets(line, sizeof(line), f);
    while (!feof(f)) {
	unsigned long iaddr;
	unsigned short iport;

	fgets(line, sizeof(line), f);
	sscanf(line, "%*d:%08lx:%04hx", &iaddr, &iport);
	if (iport != sport)
	    continue;
	if (memcmp(server, &iaddr, sizeof(iaddr)))
	    continue;
	ret = iport;
	break;
	if (verbose > 1) {
	    struct in_addr ia;
	    ia.s_addr = iaddr;
	    myprint("%slistening on %s:%d\n", ret ? "" : "not ", inet_ntoa(ia), iport);
	}
    }
    fclose(f);
    return ret;
}

static void dump(char *buf, int l, char *title)
{
    char *p = buf;
    int i = 0;
    int j;

    myprint("%sbuffer (%d bytes):\n", title, l);
    while (i < l) {
	for (j = 0; (j < 16) && ((i + j) < l); j++, p++)
	    myprint("%02x ", *p & 255);
	p -= 16;
	myprint(" ");
	for (j = 0; (j < 16) && ((i + j) < l); j++, p++)
	    myprint("%c", isprint(*p) ? *p : '.');
	myprint("\n");
	i += 16;
    }
}

typedef struct {
    struct iphdr ih;
    struct udphdr uh;
    unsigned char data[0];
} udp_packet;

static int build_packet(unsigned char *in, int len, unsigned char *out,
			struct in_addr *saddr, int sport, struct sockaddr_in *dst)
{
    udp_packet *pkt = (udp_packet *) out;
    memcpy(pkt->data, in, len);
    pkt->uh.source = ntohs(sport);
    pkt->uh.dest = dst->sin_port;
    pkt->uh.len = ntohs(8 + len);
    pkt->uh.check = 0;
    pkt->ih.version = 4;
    pkt->ih.ihl = 5;
    pkt->ih.tos = 0x10;
    pkt->ih.tot_len = 0;	/* filled by kernel */
    pkt->ih.id = 0;		/* filled by kernel */
    pkt->ih.frag_off = 0 ? 0 : 0;
    pkt->ih.ttl = IPDEFTTL;
    pkt->ih.protocol = IPPROTO_UDP;
    pkt->ih.check = 0;		/* filled by kernel */
    pkt->ih.saddr = saddr->s_addr;
    pkt->ih.daddr = dst->sin_addr.s_addr;
    return len + 28;
}

static int check_reply(const unsigned char *buf, int len, query_t *q)
{
    int i;

    if (q->reply_check_offset >= 0) {
	if (len < (q->reply_check_offset + q->reply_check_len))
	    return 0;
	return (memcmp(buf + q->reply_check_offset, q->reply,
		    q->reply_check_len) == 0);
    }
    for (i = 0; i < len - q->reply_check_len; i++) {
	if (memcmp(buf + i, q->reply, q->reply_check_len) == 0)
	    return 1;
    }
    return 0;
}

static void dispatch(struct sockaddr *client, listener_t * ll, int len, query_t *q)
{
    struct if_nameindex *iflist;
    struct sockaddr_in sin;
    int s;
    int pa;
    char reply_chk = 0;
    int bound = 0;
    int tport = ll->port;

    if (verbose > 1)
	myprint("Dispatch called\n");
    s = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if (s == -1) {
	myperror("udp socket");
	exit(-1);
    }
    sin.sin_family = AF_INET;
    sin.sin_port = htons(44444);
    iflist = if_nameindex();
    while (iflist && iflist->if_name) {
	struct ifreq ir;
	struct in_addr ia;

	if (verbose > 1)
	    myprint("Checking interface %s\n", iflist->if_name);
	strcpy(ir.ifr_name, iflist->if_name);
	if (ioctl(s, SIOCGIFADDR, &ir)) {
	    myperror("SIOCGIFADDR");
	    exit(-1);
	}
	memcpy(&ia, &((struct sockaddr_in *) &ir.ifr_addr)->sin_addr, sizeof(ia));
	if (verbose > 1)
	    myprint("Interface %s has IP of %s\n", iflist->if_name, inet_ntoa(ia));
	if ((pa = is_listening(&ia, tport))) {
	    int l;
	    int reply_sock;
	    int from_len;
	    struct sockaddr to;
	    struct sockaddr from;
	    fd_set fds;
	    struct timeval tv;
	    unsigned char reply[4096];
	    unsigned char reply_buf[4096];

	    if (verbose)
		myprint("Forwarding to %s:%d\n", inet_ntoa(ia), tport);
	    if (!bound) {
		memcpy(&sin.sin_addr, &ia, sizeof(ia));
		if (bind(s, (struct sockaddr *) &sin, sizeof(struct sockaddr))) {
		    myperror("dispatch: bind");
		    exit(-1);
		}
		bound = 1;
	    }
	    ((struct sockaddr_in *) &to)->sin_family = AF_INET;
	    ((struct sockaddr_in *) &to)->sin_port = htons(pa);
	    memcpy(&((struct sockaddr_in *) &to)->sin_addr, &ia, sizeof(ia));
	    l = sendto(s, q->query, q->query_len, 0, &to, sizeof(to));
	    if (l < 0) {
		myperror("sendto");
		exit(-1);
	    }
	    FD_ZERO(&fds);
	    FD_SET(s, &fds);
	    tv.tv_sec = 1;
	    tv.tv_usec = 0;
	    switch (select(s + 1, &fds, NULL, NULL, &tv)) {
		case 0:
		    if (verbose)
			myprint("Server reply timed out\n");
		    break;
		case 1:
		    l = recvfrom(s, reply, sizeof(reply), 0, &from, &from_len);
		    if (l < 0) {
			myperror("recvfrom");
			exit(-1);
		    }
		    if (verbose > 1)
			dump(reply, l, "reply ");
		    if ((l > 5) && check_reply(reply, l, q)) {
			if (verbose)
			    myprint("Got reply from server, forwarding back to client\n");

			reply_sock = socket(PF_INET, SOCK_RAW, IPPROTO_RAW);
			l = build_packet(reply, l, reply_buf, &ia, pa, (struct sockaddr_in *) client);
			sendto(reply_sock, reply_buf, l, 0, client, sizeof(struct sockaddr));
			close(reply_sock);
		    } else
			myprint("Got invalid response\n");
		    break;
		case -1:
		    myperror("select");
		    exit(-1);
		    break;
	    }
	}
	iflist++;
    }

    close(s);
}

static struct option opts[] = {
    {"verbose", no_argument, 0, 'v'},
    {"help", no_argument, 0, 'h'},
    {NULL, 0, 0, 0},
};

static int usage()
{
    fprintf(stderr,
	    "Usage: lanwon OPTIONS\n"
	    "\n"
	    "Supported options:\n"
	    "\n" " -h, --help           Show this text.\n" " -v, --verbose        Increase verbosity\n" "\n");
    exit(0);
}

listener_t *listeners;

static void daemonize()
{
    int i;

    signal(SIGHUP, SIG_IGN);
    signal(SIGALRM, SIG_IGN);
    if (fork())
	exit(0);
    setsid();
    for (i = 0; i < 255; i++)
	close(i);
    openlog("lanwon", LOG_PID, LOG_DAEMON);
    daemonized = 1;
}

static listener_t *bindport(int port)
{
    in_addr_t ia = INADDR_BROADCAST;
    listener_t *l = (listener_t *) malloc(sizeof(listener_t));
    int one = 1;
    struct sockaddr_in sin;

    if (!l) {
	myperror("out of memory");
	exit(-1);
    }
    l->s = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if (l->s == -1) {
	myperror("udp socket");
	exit(-1);
    }
    setsockopt(l->s, SOL_SOCKET, SO_BROADCAST, &one, sizeof(int));
    sin.sin_family = AF_INET;
    sin.sin_port = htons(port);
    memcpy(&sin.sin_addr, &ia, sizeof(ia));
    if (bind(l->s, (struct sockaddr *) &sin, sizeof(struct sockaddr))) {
	if (errno == EADDRINUSE) {
		close(l->s);
		free(l);
		myprint("port %d already in use, skipping\n", port);
		return NULL;
	} else {
		char ebuf[30];
		sprintf(ebuf, "bind on port %d", port);
		myperror(ebuf);
		exit(-1);
	}
    }
    l->port = port;
    l->next = listeners;
    l->queries = NULL;
    listeners = l;
    return l;
}

static void add_query(listener_t *l, query_t *q) {
    query_t *qq;
    query_t **lq = &(l->queries);

    while (*lq)
	lq = &((*lq)->next);
    qq = *lq = malloc(sizeof(query_t));
    if (!qq) {
	myperror("Can not allocate query");
	exit(-1);
    }
    memcpy(qq, q, sizeof(query_t));
}

int main(int argc, char **argv)
{
    struct sockaddr_in from;
    listener_t *lst;
    fd_set lfds;
    int port;
    int from_len;
    int l;
    int c;
    unsigned char buf[4096];


    while (1) {
	if ((c = getopt_long(argc, argv, "hvp:", opts, NULL)) == -1)
	    break;
	switch (c) {
	    case 'h':
		usage();
		break;
	    case 'v':
		verbose++;
		break;
	}
    }
    if (optind < argc) {
	fprintf(stderr, "No non-option arguments allowed\n");
	fprintf(stderr, "Use --help for more information\n");
	exit(1);
    }
    if (verbose < 2)
	daemonize();
    /* HLDS standard ports */
    for (port = 27015; port < 27025; port++)
	if ((lst = bindport(port))) {
	    add_query(lst, &hlds_infostring);
	    add_query(lst, &hlds_details);
	}
    /* RTCW standard ports */
    for (port = 27960; port < 27963; port++)
	if ((lst = bindport(port)))
	    add_query(lst, &q3_getinfo);
    /* JK2 standard ports */
    for (port = 28070; port < 28074; port++)
	if ((lst = bindport(port)))
	    add_query(lst, &q3_getinfo);
    /* UT standard ports */
    for (port = 8777; port < 8787; port++)
	if ((lst = bindport(port)))
	    add_query(lst, &ut_query);
    /* UT 2003 standard ports */
    for (port = 10777; port < 10787; port++)
	if ((lst = bindport(port)))
	    add_query(lst, &ut2003_query);
    /* BF1942 standard ports */
    for (port = 22000; port < 22010; port++)
	if ((lst = bindport(port)))
	    add_query(lst, &bf1942_query);

    if (verbose)
	myprint("lanwon started\n");
    while (1) {
	listener_t *lp = listeners;
	int maxfds = 0;

	FD_ZERO(&lfds);
	while (lp) {
	    FD_SET(lp->s, &lfds);
	    if (lp->s > maxfds)
		maxfds = lp->s;
	    lp = lp->next;
	}
	switch (select(maxfds + 1, &lfds, NULL, NULL, NULL)) {
	    case 0:
		break;
	    case -1:
		break;
	    default:
		lp = listeners;
		while (lp) {
		    if (FD_ISSET(lp->s, &lfds)) {
			query_t *q;

			l = recvfrom(lp->s, buf, sizeof(buf), 0,
				(struct sockaddr *)&from, &from_len);
			if (verbose > 1)
			    myprint("Got broadcast packet of %d bytes\n", l);
			q = lp->queries;
			while (q) {
			    if (memcmp(buf, q->query, q->query_len) == 0) {
				struct in_addr ia;
				memcpy(&ia, &from.sin_addr, sizeof(ia));
				if (verbose)
				    myprint("got %s query from %s\n",
					    q->name, inet_ntoa(ia));
				dispatch((struct sockaddr *) &from, lp, l, q);
				break;
			    }
			    q = q->next;
			}
		    }
		    lp = lp->next;
		}
	}
    }
    return 0;
}
